# (c) 2023 Hermann Paul von Borries
# MIT License
# freezefs file extract driver for MicroPython

import os
import errno
import sys

class _VerbosePrint:
    def __init__( self, module_name, function, silent ):
        self.silent = silent
        self.prefix = f"{module_name}.{function}"

    def print( self, *args ):
        if not self.silent:
            print( self.prefix, *args )

def _file_exists( filename ):
    try:
        open( filename ).close()
        return True
    except:
        return False
 

def _extract_file( dir_entry, destination ):
    
    from io import BytesIO
    data = dir_entry[0]
    if not data:
        return
    # Process in small chunks to reduce memory use
    buffer = bytearray(256)
    stream = BytesIO( data )
    if dir_entry[1]:
        # This is a compressed file
        # Import here, so this is imported only when needed
        from deflate import DeflateIO, AUTO
        stream = DeflateIO( stream, AUTO, 0, True )
    with open( destination, "wb" ) as outfile:
        while True:
            n = stream.readinto( buffer )
            if n == 0:
                break
            outfile.write( memoryview( buffer )[0:n] )
    stream.close()

def _create_folder( folder, vp ):   
    path = ""
    # Create parent folders first, then the specified folder
    for p in folder.split("/"):
        path += "/" + p
        path = path.replace("//", "/" )
        try:
            os.mkdir( path )
            vp.print( f"folder {path} created." )
        except Exception as e:
            if type(e) is not OSError or e.errno != errno.EEXIST:
                vp.print( f"folder {path} not created: ", e )
     
def _extract_all( direntries, target, overwrite, vp ):
    _create_folder( target, vp )
    
    # Don't mount the frozen file system,
    # access file data through the internal file structure.
    for filename, dir_entry in direntries:
        # get destination filename
        dest = (target + filename).replace("//", "/")
        if dir_entry:
            # Copy file
            try:
                if overwrite == "never" and _file_exists( dest ):
                    vp.print( f"file {dest} exists, not extracted." )
                else:
                    vp.print(  f"extracting file {dest}." )
                    _extract_file( dir_entry, dest )
            except Exception as e:
                vp.print( f"file {dest} not copied: {e}." )
                raise
        else:
            # Create folder (and it's parent folders if not created yet)
            # This relies on directory entries being sorted, so
            # the dir_entry for the folder comes up before the files it contains.
            _create_folder( dest, vp )
 
   
# Called from the generated (frozen) .py module
def extract_fs( module_name, target, overwrite, silent ):
    
    # Delete this module from list of loaded modules to help
    # free the memory after use. Calling program must use __import__
    # __main__ is not in sys.modules[]
    module = __import__( module_name )
    if module_name != "__main__":
        del sys.modules[ module_name ]
    
    vp = _VerbosePrint( module_name, "extract", silent )

    vp.print( f"extracting files to {target}." )
    _extract_all( direntries, target, overwrite, vp )
            
    return 


# (c) 2023 Hermann Paul von Borries
# MIT License


# Test path: /./folder /folder/../folder /..

import sys
import os
import unittest
import time
import errno
import gc
import shutil # mpremote mip install shutil (standard from micropython-lib)

PYTHON = sys.implementation.name
if PYTHON == "cpython":
    from pathlib import Path
else:
    Path = lambda *args : "".join( args )

referencefolder = "testfiles"
testfolder = "fz"

test_folders = [ "sub1", "sub2", "sub1/sub2", "sub1/sub2/sub3" ]

def iter_both_test_folders():
    for folder in test_folders:
        yield Path( referencefolder + "/" + folder ), Path( testfolder + "/" + folder )

# txt files are ascii, bin are binary, uni are unicode
test_files = [
    ("file1.txt", 100),
    ("file2.txt", 0),
    ("file3.bin", 100),
    ("file4.bin", 0),
    ("file5.txt", 1), 
    ("file6.uni", 10), # Size of unicode files is in characters, not bytes
    ("file7.uni", 100), # Size of unicode files is in characters, not bytes
    ("file8.uni", 1000), # Size of unicode files is in characters, not bytes
    ("file9.uni", 10000), # Size of unicode files is in characters, not bytes
    ("sub1/file1.txt", 10),
    ("sub1/file2.txt", 20),
    ("sub1/sub2/file1.txt", 13),
    ("sub1/sub2/file2.txt", 11 ),
    ("sub1/sub2/file3.txt", 21 ),
    ("sub1/sub2/sub3/file4.txt", 77 )
]

        
def iter_both_test_files( filetype=None ):
    for f, length in test_files:
        if filetype == None or f.endswith( "." + filetype ):
            yield Path( referencefolder + "/" + f ), Path( testfolder + "/" + f )


# Old fashioned pseudorandom generator to generate
# same test data on PC and microcontroller
rand = 11
m = 2**31-1
a = 1664525
c = 1013904223
def getrand(n):
    global rand
    rand += (a*rand+c)%m
    return rand%n


# Some characters for testing
unicodes = "aáéíóúÁÉÍÓÚäëïöüÄËÏÖÜñÑ" + chr(0x1f600) + chr(0x1f603) + chr(0x1f604)
asciis = "".join( chr(x) for x in range(0,127) ) + "\r" + "\n" + "\t"

def write_str( filename, length, alphabet ):
    # Don't use newline to mirror Micropython
    if PYTHON == "cpython":
        fileopen = lambda fn : open( filename, "wt", encoding="utf-8", newline="" )
    else:
        fileopen = lambda fn : open( filename, "wt" )
    with fileopen( filename ) as file:
        for _ in range(length):
            p = getrand( len(alphabet) )
            s = alphabet[p:p+1]
            file.write( s )

def write_text( filename, length ):
    write_str( filename, length, asciis ) 

def write_bin( filename, length ):
    with open( filename, "wb" ) as file:
        for _ in range(length):
            b = bytearray(1)
            b[0] = getrand( 255 ) 
            file.write( b )

       
def write_uni( filename, length ):
    if length <= 10:
        write_str( filename, length, unicodes )
    else:
        write_str( filename, length, asciis + unicodes )

def make_testfiles( ): 
    try:
        shutil.rmtree( referencefolder )
    except:
        pass
    # This has to run both on PC/Python and MicroPython
    # Create folders
    for f in [""] + test_folders:
        if f == "":
            filename = referencefolder
        else:
            filename = referencefolder + "/" + f
        print("make_testfiles: Create folder", filename )
        try:
            os.mkdir( Path( filename ) )
        except Exception as e:
            print(f"make_testfiles: Could not create folder {filename}, {e}" )
            # Ignore errors, sometimes delete folder fails on Windows
     
    # Write files
    for f, length in test_files:
        filename = Path( referencefolder + "/" + f )
        if ".txt" in f:
            write_text( filename, length )
        elif ".bin" in f:
            write_bin( filename, length )
        elif ".uni" in f:
            write_uni( filename, length )
    print("make_testfiles: Test files created in folder ", referencefolder)

    
class TestFiles(unittest.TestCase):

    def compare_lists( self, list1, list2 ):
        # With infinite memory, this could compare tuple( list1 ) with tuple( list2 )
        self.assertEqual( len(list1), len(list2) )
        for i, p in enumerate( list1 ):
            self.assertEqual( p, list2[i] )
         
         
    def check_testfile_size( self, folder ):
        for f, defsize in test_files:
            if ".uni" in f:
                # File size of unicode files is in characters, not bytes
                # can't compare here
                continue
            filename = "/" + folder + "/" + f
            filename = filename.replace("//", "/")
            try:
                size = os.stat( filename )[6]
            except OSError:
                break
            self.assertEqual( defsize, size )

    def test_full_read_binary( self ):
        for filename1, filename2 in iter_both_test_files( "bin" ):
            with open( filename1, "rb" ) as file1:
                c1 = file1.read()
            with open( filename2, "rb") as file2:
                c2 = file2.read()
            self.assertEqual( c1, c2 )

    def test_full_read_text( self ):
        for filename1, filename2 in iter_both_test_files( "txt" ):
            with open( filename1, "r" ) as file1:
                c1 = file1.read()
            with open( filename2, "r" ) as file2:
                c2 = file2.read()
            
            self.assertEqual( c1, c2 )

    def test_full_try_binary_as_text( self ):
        for filename1, filename2 in iter_both_test_files( "bin" ):
            if os.stat( filename1 )[6] == 0 and os.stat( filename2 )[6] == 0:
            # Length 0, will not raise error
                continue
            # Reading binary file as text should raise UnicodeError
            with self.assertRaises( UnicodeError ):
                with open( filename1, "r" ) as file1:
                    c1 = file1.read()
            with self.assertRaises( UnicodeError ):
                with open( filename2, "r") as file2:
                    c2 = file2.read()
           

    def test_full_read_0( self ):
        for file1, file2 in iter_both_test_files( "txt" ):
            with open( file1, "r" ) as file:
                c1 = file.read(0)
                d1 = file.read(10)
                e1 = file.read(0)
                f1 = file.read(20)
            with open( file2, "r" ) as file:
                c2 = file.read(0)
                d2 = file.read(10)
                e2 = file.read(0)
                f2 = file.read(20)
            
            self.assertEqual( c1, c2 )
            self.assertEqual( d1, d2 )
            self.assertEqual( e1, e2 )
            self.assertEqual( f1, f2 )

    
    def test_read_entire_text( self ):
        print("")
        for filetype in ("txt", "uni"):
            for filename1, filename2 in iter_both_test_files( filetype ):
                
                with open( filename1, "r") as file1:
                    gc.collect()
                    data1 = file1.read()
                    print(f"test_read_entire_text {filename1=} {len(data1)=} characters")
                with open( filename2, "r") as file2:
                    gc.collect()
                    data2 = file2.read()
                    print(f"test_read_entire_text {filename2=} {len(data2)=} characters")
                self.assertEqual( data1, data2 )
                data1 = None
                data2 = None
                gc.collect()
                
    def test_read_text( self ):
        for filetype in ("txt", "uni"):
            for length in ( 1, 10, 100, 1000 ):
                print(f"Test read text with read({length})" )
                for filename1, filename2 in iter_both_test_files( filetype ):
                    file1 = open( filename1, "r" )
                    file2 = open( filename2, "r" )
                    while True:
                        data1 = file1.read( length )
                        data2 = file2.read( length )
                        self.assertEqual( data1, data2 )
                        if not data1:
                            break
                    file1.close()
                    file2.close()


    def test_readline( self ):
        for filetype in ("txt", "uni"):
            for filename1, filename2 in iter_both_test_files( filetype ):
                for mode in ("r", "rb"):
                    file1 = open( filename1, mode )
                    file2 = open( filename2, mode )
                    while True:
                        data1 = file1.readline()
                        data2 = file2.readline()
                        self.assertEqual( data1, data2 )
                        if not data1:
                            break
                    file1.close()
                    file2.close()

               

    def test_readline3( self ):
        for filetype in ("txt", "uni"):
            for file1, file2 in iter_both_test_files( filetype ):
                with open( file1, "r" ) as file:
                    parts1 = [ _ for _ in file ]
                with open( file2, "r") as file:
                    parts2 = [ _ for _ in file ]
                self.compare_lists( parts1, parts2 )


    def test_read_binary( self ):
        for length in ( 1, 10, 100, 1000 ):
            print(f"Test read binary with read({length})" )
            for filename1, filename2 in iter_both_test_files( "bin" ):
                file1 = open( filename1, "rb" )
                file2 = open( filename2, "rb" )
                while True:
                    data1 = file1.read( length )
                    data2 = file2.read( length )
                    self.assertEqual( data1, data2 )
                    if not data2:
                        break
                file1.close()
                file2.close()
 
    def test_listdir( self ):
        for rf, tf in iter_both_test_folders():
            files1 = os.listdir( "/" + rf )
            files2 = os.listdir( "/" + tf )
            self.assertEqual( tuple(files1), tuple(files2) )
    
    def test_chdir( self ):
        for fol in test_folders:
            os.chdir( "/" + referencefolder + "/" + fol )
            files1 = os.listdir( "" )
            os.chdir( "/" + testfolder + "/" + fol )
            files2 = os.listdir( "" )
            self.assertEqual( tuple( files1 ), tuple( files2 ) )

        # chdir with ending /
        os.chdir( "/" + referencefolder + "/sub1/"  )
        self.assertEqual( "/" + referencefolder + "/sub1", os.getcwd() )
        os.stat( "file9.uni" )
        files1 = os.listdir( "" )
        
        
        os.chdir( "/" + testfolder + "/sub1/" )
        self.assertEqual( "/" + testfolder + "/sub1", os.getcwd() )
        os.stat( "file9.uni" )
        files2 = os.listdir( "" )
  
        self.assertEqual( tuple( files1 ), tuple( files2 ) )

        
        # Bad chdir: to nonexistent folder
        for folder in ( referencefolder, testfolder ):
            with self.assertRaises( OSError ):
                os.chdir("/" + folder + "/nonexistent")

        # Bad chdir: to file
        for folder in ( referencefolder, testfolder ):
            with self.assertRaises( OSError ):
                os.chdir("/" + folder + "/file1.txt")

        # chdir to relative path
        for folder in ( referencefolder, testfolder ):
            newdir = "/" + folder
            os.chdir( newdir )
            self.assertEqual( mewdir, os.getcwd() )
            newdir += "sub1"
            os.chdir( "sub1" )
            self.assertEqual( mewdir, os.getcwd() )
            newdir += "sub2"
            os.chdir( "sub2" )
            self.assertEqual( mewdir, os.getcwd() )

        os.chdir(rf + "/sub1/sub2/.." )
        files1 = os.listdir()
        os.chdir( tf + "/sub1/sub2/.." )
        files2 = os.listdir()
        self.assertEqual(files1, files2 )
        
        os.chdir(rf + "/sub1/sub2/." )
        files1 = os.listdir()
        os.chdir( tf + "/sub1/sub2/." )
        files2 = os.listdir()
        self.assertEqual(files1, files2 )

        os.chdir(rf + "/sub1/./sub2/.." )
        files1 = os.listdir()
        os.chdir( tf + "/sub1/./sub2/.." )
        files2 = os.listdir()
        self.assertEqual(files1, files2 )
        
        os.chdir(rf + "/,/sub1/sub2/.." )
        files1 = os.listdir()
        os.chdir( tf + "/./sub1/sub2/.." )
        files2 = os.listdir()
        self.assertEqual(files1, files2 )

        
        os.chdir("/")
        

    def test_ilistdir( self ):
        for rf, tf in iter_both_test_folders():
            files1 = [ _ for _ in os.listdir( "/" + rf ) ]
            files2 = [ _ for _ in os.listdir( "/" + tf ) ]
            self.assertEqual( tuple(files1), tuple(files2) )

        
    def test_stat( self ):
        for filename, length in test_files:
            file1 = "/" + testfolder + "/" + filename
            file2 = "/" + referencefolder + "/" + filename
            stat1 = os.stat( file1 )[0:6]
            stat2 = os.stat( file2 )[0:6]
        self.assertEqual( stat1, stat2 )

        stat1 = os.stat( testfolder )[0:6]
        stat2 = os.stat( referencefolder )[0:6]
        self.assertEqual( stat1, stat2 )
    
    def test_parent_dir( self ):
        rf = "/" + referencefolder
        tf = "/" + testfolder
        stat1 = os.stat( rf + "/sub1/sub2/../file1.txt" )
        stat2 = os.stat( tf + "/sub1/sub2/../file1.txt" )
        self.assertEqual( stat1[0:6], stat2[0:6] )
        
        os.chdir( rf + "/sub1/../sub1/sub2")
        p1 = os.getcwd().replace( rf, "/<root>")
        os.chdir( tf + "/sub1/../sub1/sub2")
        p2 = os.getcwd().replace( tf, "/<root>")
        self.assertEqual( p1, p2 )
        
        with self.assertRaises( OSError ):
            os.chdir( tf + "/.." )


        os.chdir("/")
        
    def test_stat_folder( self ):
        for filename in test_folders:
            file1 = "/" + testfolder + "/" + filename
            file2 = "/" + referencefolder + "/" + filename
            stat1 = os.stat( file1 )[0:6]
            stat2 = os.stat( file2 )[0:6]
        self.assertEqual( stat1, stat2 )

    def test_chdir( self ):
        tf = "/" + testfolder
        rf = "/" + referencefolder
        os.chdir( tf )
        self.assertEqual( os.getcwd(), tf )
        files1 = os.listdir("")
        files2 = os.listdir( rf )
        self.assertEqual( tuple( files1 ), tuple( files2 ) )

        os.chdir("sub1")
        self.assertEqual( os.getcwd(), tf +"/sub1")
        files1 = os.listdir("")
        files2 = os.listdir( rf + "/sub1" )
        self.assertEqual( tuple( files1 ), tuple( files2 ) )

        os.chdir( tf +"/sub1")
        self.assertEqual( os.getcwd(), tf + "/sub1")
        files1 = os.listdir("")
        files2 = os.listdir( rf  + "/sub1" )
        self.assertEqual( tuple( files1 ), tuple( files2 ) )
        
        os.chdir("sub2")
        self.assertEqual( os.getcwd(), "/fz/sub1/sub2")
        files1 = os.listdir("")
        files2 = os.listdir( rf  + "/sub1/sub2" )
        self.assertEqual( tuple( files1 ), tuple( files2 ) )

        os.chdir("sub3")
        self.assertEqual( os.getcwd(), "/fz/sub1/sub2/sub3")
        files1 = os.listdir("")
        files2 = os.listdir( rf  + "/sub1/sub2/sub3" )
        self.assertEqual( tuple( files1 ), tuple( files2 ) )
        
        os.chdir( "/fz" )
        self.assertEqual( os.getcwd(), "/fz")
        files1 = os.listdir("")
        files2 = os.listdir( "/" + referencefolder  )
        self.assertEqual( tuple( files1 ), tuple( files2 ) )
        
        os.chdir("/")
    
    def test_remove( self ):
        with self.assertRaises( OSError ):
            os.remove( testfolder +"/file1.txt")
    
    def test_write_not_allowed( self ):
        file1 = testfolder +"/file1.txt"
        with self.assertRaises( OSError ):
            open( file1, "w")
        with self.assertRaises( OSError ):
            open( file1, "wb")
        with self.assertRaises( OSError ):
            open( file1, "r+")
        with self.assertRaises( OSError ):
            open( file1, "a")
            
        with self.assertRaises( OSError ):
            os.rename( file1, testfolder + "/a.a")
            
        with self.assertRaises( OSError ):
            os.remove( file1 )
            

    def test_seek_binary( self ):
        # Test seek/tell on binary files
        file1 = open( referencefolder + "/file3.bin" ,"rb")
        file2 = open( testfolder + "/file3.bin", "rb" )
        for i in [1,51,2,92]:
            file1.seek(i)
            file2.seek(i)
            self.assertEqual( file1.tell(), file2.tell() )
            c1 = file1.read(1)
            c2 = file2.read(1)
            self.assertEqual( c1, c2 )

        for i in [1,51,-29,33]:
            file1.seek(i,1)
            file2.seek(i,1)
            self.assertEqual( file1.tell(), file2.tell() )
            c1 = file1.read(1)
            c2 = file2.read(1)
            self.assertEqual( c1, c2 )

        file1.seek(i,2)
        file2.seek(i,2)
        self.assertEqual( file1.tell(), file2.tell() )
        c1 = file1.read(1)
        c2 = file2.read(1)
        self.assertEqual( c1, c2 )
 
    def test_flush( self ):
        file1 = open( "/" + referencefolder + "/file3.bin" ,"rb")
        file2 = open( "/" + testfolder + "/file3.bin", "rb" )
        self.assertEqual( file1.flush(), file2.flush() )
     
    def test_statvfs( self ):
        t = os.statvfs( "/" + testfolder )
        self.assertEqual( (1, 1, 13578, 0, 0, 19, 0, 0, 1, 255), t )


 
def timing_tests():
    print("\nTiming tests. /testfolder is on the standard file system, /fz is the VfsFrozen file system running in RAM.\n" )  
    class HowLong:
        def __init__( self, name ):
            self.name = name
        def __enter__( self ):
            self.t0 = time.ticks_ms()
            return self
        def __exit__( self, exception_type, exception_value, traceback ):
            print(self.name, time.ticks_diff( time.ticks_ms(), self.t0 ))

    size = os.stat( referencefolder + "/file9.uni" )[6]
    
    for mode in ("r", "rb"):
        for i in (1,10,100,1000):
            for folder in (referencefolder, testfolder ) :
                filename = folder + "/file9.uni"
                size = os.stat( filename )[6]
                with HowLong(f"Timed file.read({i:4d})  for {filename:19s}, size={size} bytes, mode={mode:2s}, msec="):
                    with open( filename, mode ) as file:
                        s = ""
                        while True:
                            r = file.read(i)
                            if len(r) == 0:
                                break
               

    for folder in (referencefolder, testfolder ) :
        filename = folder + "/file9.uni"
        with HowLong(f"Timed file.readline()  for {filename:19s}, size={size} bytes, mode=rb, msec="):
            with open( filename, "rb") as file:
                while True:
                    line = file.readline()
                    if len(line) == 0:
                        break



 
if __name__ == "__main__":
    if PYTHON == "cpython":
        make_testfiles()
        sys.exit()
    gc.collect()
    # Collect frequently
    gc.threshold( gc.mem_free()//2)
    make_testfiles()
    
    import frozenfiles_mount
    
    unittest.main()
    
    # Run without unittest:
    #t = TestFiles() 
    #t.test_chdir()
    
    timing_tests()
    



# (c) 2023 Hermann Paul von Borries
# MIT License
# MicroPython VFS mount driver for freezefs
import os
import errno
from io import BytesIO, StringIO
from collections import OrderedDict
import sys

def _get_basename( filename ):
    return filename.split("/")[-1]

def _get_folder( filename ):
    basename = _get_basename( filename )
    folder = filename[0:-len(basename)-1]
    if folder == "":
        folder = "/"
    return folder

 
class VfsFrozen:
    # File system for frozen files. Implements mount, listdir, stat, open etc
    # Actual read, readinto, seek, are done by BytesIO and StringIO
    def __init__( self, direntries, sum_size, files_folders ):       
        # direntries is (filename, ( data, compressed, size ) for files
        #               (filename, None) for folders
        self.filedict = OrderedDict( direntries )
        self.sum_size = sum_size
        self.files_folders = files_folders
        self.path = "/"
    
    def _to_absolute_filename( self, filename ):
        if not filename.startswith( "/" ) :
            filename = self.path + "/" + filename
        if filename.endswith("/") and filename != "/":
            filename = filename[0:-1]
        filename = filename.replace("//", "/")

        # Solve ".."
        parts = filename.split("/")
        i = -1
        while ".." in parts:
            i = parts.index("..")
            if i > 1:
                del parts[i]
                del parts[i-1]
            else:
                # Can't access /..
                self._raise_perm()
        # Solve ./file or /folder/./file
        while "." in parts:
            i = parts.index(".")
            del parts[i]
        filename = "".join( "/" + p for p in parts ).replace( "//", "/" )
        return filename
        
    def _find_file( self, filename ):
        filename = self._to_absolute_filename( filename )
        if filename == "/":
            # The root is a folder. filedict doesn't have 
            # the root. Return folder direntry for the root.
            return None
        if filename in self.filedict:
            # Return the file directory entry (dir_entry)
            return self.filedict[filename]
        else:
            raise OSError( errno.ENOENT )
    
    def _raise_perm( self ):
        raise OSError(errno.EPERM) # Very common here
        
    def open( self, filename, mode, buffering=None ):
        # Validate mode before opening file
        for c in mode:
            # Modes may be "r"/"rt"/"tr" or "rb"/"br"
            if c not in "rbt":
                raise OSError( errno.EINVAL )
            
        dir_entry = self._find_file( filename )
        if dir_entry is None: 
            # This is a folder or the root of this file system
            if filename == "/":
                self._raise_perm()
            raise OSError(errno.EISDIR)
            
        data = dir_entry[0] # data 
        if not dir_entry[1]: # compressed
            if "b" in mode:
                return BytesIO( data )
            else:
                return StringIO( data )
        else:
            # Compressed file - late import of deflate library
            from deflate import DeflateIO, AUTO
            uncompressed_stream = DeflateIO( BytesIO( data ),
                            AUTO, 0, True )
            if "b" in mode:
                return uncompressed_stream
            else:
                # This requires to buffer the entire file...
                # Only useful if enough RAM is available.
                return StringIO( uncompressed_stream.read() )
        
    def chdir( self, path ):
        newdir = self._to_absolute_filename( path )
        dir_entry = self._find_file( newdir )
        if dir_entry is None:
            # ok, it's a folder
            self.path = newdir
            return
        raise OSError( -2 )
        
    def getcwd( self ):
        if self.path != "/" and self.path.endswith( "/" ):
            return self.path[0:-1]
        return self.path
        
    def ilistdir( self, path ):
        abspath = self._to_absolute_filename( path ) 
        # Test if folder exists, if not raise OSError ENOENT
        self._find_file( abspath )
        # Find all files
        for filename, dir_entry in self.filedict.items():
            if _get_folder( filename ) == abspath:
                basename = _get_basename( filename )
                if dir_entry is not None:
                    yield ( basename, 0x8000, 0,  dir_entry[2] )
                else:
                    yield ( basename, 0x4000, 0, 0 )
                                         
    def stat( self, filename ): 
        dir_entry = self._find_file( filename )
        if dir_entry is None:
            return (0x4000, 0,0,0,0,0, 0, 0,0,0)
        else:
            return (0x8000, 0,0,0,0,0, dir_entry[2], 0,0,0)
        
    
    def statvfs( self, *args ):
        # statvfs returns:
        # block size, fragment size, blocks, 
        # free blocks, available blocks, files,
        # free inodes1, free_inodes2, mount flags=readonly
        # maximum filename length
        # Return block size of 1 byte = allocation unit
        # No free space. One "inode" per file or folder. 
        # Mount flags: readonly
        # Max filename size 255 (checked in freezeFS)
        #sum_size = sum( d[2] for d in self.filedict.values() if d is not None )
        return (1,1,self.sum_size,  0,0,self.files_folders,  0,0,1, 255)
    
    def mount( self, readonly, x ):
        self.path = "/"
        
    def remove( self, filename ):
        self._raise_perm()
    
    def mkdir( self, *args ):
        self._raise_perm()

    def rename( self, oldfname, newfname ):
        self._raise_perm()
        
    def umount( self ):
        # No specific cleanup necessary on umount.
        pass
    
    # Added rmdir for completeness
    def rmdir(self):
        self._raise_perm()
        
def mount_fs( frozen_module_name, target, silent ):
    
    if target is None:
        raise ValueError("No target specified")

    # Check target doesn't exist
    file_exists = False
    try:
        os.stat( target )
        file_exists = True
    except:
        pass  
    if file_exists:
        raise OSError( errno.EEXIST )
    
    if not silent:
        print( f"mounting {__name__} at {target}." )
        
    os.mount( VfsFrozen( direntries, sum_size, files_folders ), target, readonly=True )
    return True



# (c) 2023 Hermann Paul von Borries
# MIT License
# freeze
import argparse
from pathlib import Path, PurePosixPath
from glob import glob
import os
import sys
import time
import zlib

MAX_FILENAME_LEN = 255
VERSION = 2 # Directory entries changed
silent = False
def _verbose_print( *args ):
    if silent:
        return
    print( *args )

varcounter = 0
class FileObject:
    def __init__( self, pc_infolder, path, request_compression, level, wbits ):
        global varcounter
        self.path = path 
        self.request_compression = request_compression
        self.level = level
        self.wbits = wbits
        
        self.mp_path = "/" + self.path.as_posix()
        if len( self.mp_path ) >= MAX_FILENAME_LEN:
            raise ValueError(f"File name too long {self.mp_path}")
        
        self.pc_path = pc_infolder / path
        
        # Assign a variable name and get the file data (raw or compressed)
        self.is_file = self.pc_path.is_file()
        if self.is_file:
            self.varname = f"_f{varcounter}"
            varcounter += 1
            self._get_data()
            
    def _get_data( self ):
        # get data, compress (if indicated)
        # assign self.size, self.compressed_size, self.compressed
        self.data = self._file_read( )
        self.size = len( self.data )
        self.compressed = False
        self.compressed_size = self.size
        if self.request_compression:
            compressed_data = self._zlib_compress( self.data )
            if len( compressed_data ) < self.size:
                # Use compression if compression gives some gain
                self.data = compressed_data
                self.compressed = True
                self.compressed_size = len( self.data )
    
    def _zlib_compress( self, original_data ):
        # This will create the zlib header (8 bytes) that contains level and wbits for deflate
        zco = zlib.compressobj( level=self.level, wbits=self.wbits )
        compressed_data = zco.compress( self.data )
        compressed_data += zco.flush()
        return compressed_data


    def get_pythonized( self ):
        pythonized = ""
        p = 0
        while True:
                chunk = self.data[p:p+16]
                pythonized += f"  {str(chunk)}\\\n"
                if len(chunk) == 0:
                    break
                p += 16
            
        # Last item without \ nor \n 
        return pythonized[0:-2]

    def _file_read( self ):
        with open( self.pc_path, "rb" ) as file:
            return file.read()
    
      
def to_python(  pc_infolder, pc_outfile,
                mc_target, on_import, overwrite,
                silent,
                request_compression, wbits, level ):
        
    # Get files
    files = []
    for path in glob( "**", root_dir=pc_infolder, recursive=True ):
        fo = FileObject( pc_infolder, Path( path ), request_compression, level, wbits )
        files.append( fo )

    # Generate output        
    files.sort( key=lambda fo: fo.mp_path )         
    with open( pc_outfile, "w", encoding="utf-8") as outfile:
        _files_to_python( outfile, files, silent )
        _generate_appended_code( outfile,
                                 mc_target, on_import, overwrite,
                                 silent )
        
    # Print some statistics
    sum_size = sum( fo.size for fo in files if fo.is_file  )
    number_of_files = sum( 1 for fo in files if fo.is_file )
    number_of_folders = len( files ) - number_of_files
    _verbose_print(f"Sum of file sizes {sum_size} bytes, {number_of_files} files {number_of_folders} folders" ) 
    if request_compression and sum_size != 0:
        sum_compressed_size = sum( fo.compressed_size for fo in files if fo.is_file )
        r = sum_compressed_size/sum_size*100
        _verbose_print(f"Sum of compressed sizes {sum_compressed_size}, compressed/original={r:.1f}%")

    return True
    
def _files_to_python( outfile, files, silent ):
        
    # Generate one const for each file
    for fo in files:
        if fo.is_file:
            pythonized = fo.get_pythonized()
            c = ""
            if fo.compressed:
                c = f", compressed level={fo.level} wbits={fo.wbits} compressed size={fo.compressed_size}" 
            outfile.write(f"# {fo.mp_path} size={fo.size}{c}\n" ) 
            outfile.write(f"{fo.varname} = const(\n{pythonized})\n")
            
    # Generate the directory entries
    outfile.write("direntries = const((")
    for fo in files:            
        if fo.is_file:
            direntry = f"( {fo.varname}, {fo.compressed}, {fo.size} )"
            ratio = ""
            if fo.compressed:
                r = fo.compressed_size / fo.size * 100
                ratio = f", compressed/original={r:.0f}%" 
            _verbose_print(f"Appended file   {fo.pc_path} ({fo.size} bytes) as {fo.mp_path}{ratio}" )
        else:
            direntry = f"None"
            _verbose_print(f"Appended folder {fo.pc_path} as {fo.mp_path}")
        outfile.write(f" ( '{fo.mp_path}',  {direntry} ),\n" )
    outfile.write("))\n\n" )
    # Generate file info
    outfile.write(f"version = const({VERSION})\n")
    t = time.localtime()        
    outfile.write(f"date_frozen = const( '{t[0]}/{t[1]:02d}/{t[2]:02d} {t[3]:02d}:{t[4]:02d}:{t[5]:02d}' )\n\n")
    sum_size = sum( fo.size for fo in files if fo.is_file )
    outfile.write(f"sum_size = const({sum_size})\n" )
    outfile.write(f"files_folders = const({len(files)})\n" )

def _generate_appended_code( outfile,
                             mc_target, on_import,  overwrite, silent ):

        # Compute which drivers to include

        include_file = "ffs" + on_import + ".py"
        # Include driver            
        
        # Open driver files same folder where freezefs.py is
        f = Path( __file__ ).parent / include_file
        with open( f, "r" ) as file:
            while True:
                line = file.readline()
                if line == "":
                    break
                # Don't copy comments nor empty lnes
                s = line.strip(" ")
                if s.startswith("#") or s == "\n":
                    continue
                # Make indents one by one instead of by four
                line = line.replace("    ", " ")
                outfile.write( line )
        outfile.write("\n")
 
        if on_import == "mount":
            outfile.write(f"mount_fs( __name__, '{mc_target}', {silent} )")
        elif on_import == "extract":
            outfile.write(f"extract_fs( __name__, '{mc_target}', '{overwrite}', {silent} )")   


DESC = """freezefs.py
freezefs saves a file structure with subfolders and builds an self-extractable or self-mountable archive in a .py file, optionally with compression, to be frozen as bytecode or extracted.


Examples:
freezefs.py input_folder frozenfiles.py --target=/myfiles --on_import mount
freezefs.py input_folder frozenfiles.py --target=/myfiles --on_import=extract --compress

"""
def main():
    parser = argparse.ArgumentParser(
                "python -m freezefs", 
                description=DESC,
                formatter_class = argparse.RawDescriptionHelpFormatter)
    parser.add_argument('infolder', type=str, help='Input folder path')
    parser.add_argument('outfile', type=str,
                        help='Path and name of output module. Must have .py extension.')

    parser.add_argument("--on-import", "-oi", type=str,
                        dest="on_import",
                        choices=["mount", "extract"],
                        default="mount",
                        help="Action when importing output module. Default is mount.")

    parser.add_argument("--target", "-t", type=str,
                        dest="target",
                        help="For --on-import=mount: mount point."
                             " For --on-import=extract: destination folder."
                             " Default: the infolder."
                             " Example: --target /myfiles. Must start with /")
                             
    parser.add_argument("--overwrite", "-ov", type=str,
                        dest="overwrite",
                        choices=["never", "always"],
                        default="never",
                        help="always: on extract, all files are overwritten. never: on extract, no file is overwritten, only new files are extracted. Default: never. ")
                        
    # Compress parameters
    parser.add_argument("--compress", "-c",
                        dest="compress", default=False,
                        action=argparse.BooleanOptionalAction,
                        help="Compress files before writing to output .py. See python zlib compression." )

    parser.add_argument("--wbits", "-w",
                        type=int,
                        dest="wbits",
                        default=10,
                        help="Compression window of 2**WBITS bytes. Between 9 and 14."
                              " Default is 10 (1024 bytes)")
    parser.add_argument("--level", "-l",
                        type=int,
                        dest="level",
                        default=9,
                        help="Compression level. Between 0 (no compression) and 9 (best compression)."
                             " Default is 9" )
    parser.add_argument("--silent", "-s", 
                        dest="silent", default=False,
                        action="store_true",
                        help="Supress messages printed when mounting/copying files"
                              " and while running this program.")

    args = parser.parse_args()
    
    # Use pathlib paths to make code independent of operati
    pc_infolder = Path( args.infolder )
    pc_outfile = Path( args.outfile )


    
    if not pc_infolder.is_dir():
        quit("Input folder does not exist, or is not a folder")

    if not pc_outfile.suffix.lower() == '.py':
        quit('Output filename must have a .py extension.')

    if args.target:
        mc_target = PurePosixPath( args.target )
        if not args.target.startswith("/"):
            quit( "Target must start with /")

        if args.target.endswith("/") and args.target != "/":
            quit( "Target must not end with /")
    else:
        mc_target = PurePosixPath( "/" + pc_infolder.stem )
        print( f"Target set to {mc_target}" )
        
    if not ( 9 <= args.wbits <= 14 ):
        quit("--wbits must be between 9 and 14")
        
    if not( 0 <= args.level <= 9 ):
        quit("--level must be between 0 and 9" )
      

    if mc_target.stem == pc_outfile.stem:
        print(f"--target must be a different name than output file: {pc_outfile.stem}")
        sys.exit(1)

    _verbose_print(f'Writing Python file {pc_outfile}.') 
    if not to_python( pc_infolder, pc_outfile,
                      mc_target, args.on_import, args.overwrite,
                      args.silent,
                      args.compress, args.wbits, args.level ):
        sys.exit(1)


    module_name = pc_outfile.stem

    if args.on_import == "mount":
        _verbose_print(f"On import the file system will be mounted at {mc_target}." )
    elif args.on_import == "extract":
        if args.overwrite == "never":
            _verbose_print(f"On import the file system will be extracted to {mc_target} writing only files that don't exist." )
        else:
            _verbose_print(f"On import the file system will be extracted to {mc_target} overwriting all files." )

    if args.on_import == "mount" and args.compress:
        # This is the only combination that is heavy in RAM use.
        print("sWarning: --on-import=mount and --compress loads file in RAM when opening files in text 'r' mode. High RAM use." )
        
    _verbose_print(pc_outfile, 'written successfully.')
